#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "main.h"
#include "Graphics.h"
#include "System.h"
#include "math/Complex.h"

#define SCREEN_WIDTH    1280
#define SCREEN_HEIGHT   720
#define MAX_ITERATIONS  1000
#define TOLERANCE       1e-6

// the function
Complex f(Complex z)
{
    // z^5 - 3iz^3 - (5+2i)z^2 + 3z + 1
    return z * z * z * z * z - Complex(0, 3) * z * z * z - Complex(5, 2) * z * z + z * 3 + Complex(1, 0);
}

// derivative of the function
Complex df(Complex z)
{
    // 5z^4 - 9iz^2 - (10+4i)z + 3
    return z * z * z * z * 5 - Complex(0, 9) * z * z - Complex(10, 4) * z + Complex(3, 0);
}

int main(int argc, char* argv[])
{
    // init the window
    gfx.init("Newton fractal 5", SCREEN_WIDTH, SCREEN_HEIGHT);
    gfx.init2D();
    gfx.clearScreen(Color(0, 0, 0, SDL_ALPHA_OPAQUE));

    static double zoom = 5.0 / (double)SCREEN_HEIGHT;
    static double centerX = (double)SCREEN_WIDTH  / 2.0;
    static double centerY = (double)SCREEN_HEIGHT / 2.0;

    // roots of the function
    Complex roots[5];
    roots[0] = Complex(-1.28992, -1.87357);
    roots[1] = Complex(-0.824853, 1.17353);
    roots[2] = Complex(-0.23744, 0.0134729);
    roots[3] = Complex(0.573868, -0.276869);
    roots[4] = Complex(1.77834,  0.963437);

    // palette
    static Color    palette[6];
    palette[0] = Color(0, 255, 200);
    palette[1] = Color(240, 0, 135);
    palette[2] = Color(200, 200, 0);
    palette[3] = Color(255, 90, 0);
    palette[4] = Color(40, 240, 40);
    palette[5] = Color(0, 0, 0);

    // draw the fractal
    for (int y = 0; y < SCREEN_HEIGHT; y++)
    {
        for (int x = 0; x < SCREEN_WIDTH; x++)
        {
            // convert the pixel pos to a complex number
            Complex z;
            z.re = (x - centerX) * zoom;
            z.im = (y - centerY) * zoom;

            // by default the pixel is black if the calculations fail
            int palIdx = 5;

            // iterate
            int iter;

            for (iter = 0; iter < MAX_ITERATIONS; iter++)
            {
                z = z - f(z) / df(z);

                // are we near a root ?
                for (int i = 0; i < 5; i++)
                {
                    Complex delta = z - roots[i];
                    double distance = delta.squareMag();

                    if (distance < TOLERANCE)
                    {
                        palIdx = i;
                        break;
                    }
                }

                if (palIdx != 5)
                    break;
            }

            // draw the pixel
            Color color = palette[palIdx];

            if (palIdx != 5)
            {
                float coef = 1 - log(iter + 1) / 5;

                if (coef < 0)
                    coef = 0;

                color.r *= coef;
                color.g *= coef;
                color.b *= coef;
            }
            gfx.setPixel(x, y, color);
        }

        gfx.render();
        sys.processEvents();

        if (sys.isQuitRequested() == true)
            break;
    }

    // wait until we quit
    while (sys.isQuitRequested() == false)
    {
        sys.processEvents();
        sys.wait(16);
    }

    gfx.quit();

    return EXIT_SUCCESS;
}
